# -*- coding: utf-8 -*-
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Data Format
    ~~~~~~~~~~~

    Base64, Base58, Hex, UTF-8, JsON
"""

import base64
import json
from abc import abstractmethod
from typing import Optional, Union


class DataCoder:

    @abstractmethod
    def encode(self, data: bytes) -> str:
        """
        Encode binary data to text string

        :param data: binary data
        :return:     text string (Base58/64, Hex, ...)
        """
        raise NotImplemented

    @abstractmethod
    def decode(self, string: str) -> Optional[bytes]:
        """
        Decode text string to binary data

        :param string: text string (Base58/64, Hex, ...)
        :return:       binary data
        """
        raise NotImplemented


class DataParser:

    @abstractmethod
    def encode(self, o: Union[str, dict, list]) -> bytes:
        """
        Encode container/string object to bytes

        :param o: Map, List, or String
        :return: JsON data or UTF-8 string bytes
        """
        raise NotImplemented

    @abstractmethod
    def decode(self, data: bytes) -> Union[str, dict, list, None]:
        """
        Decode bytes to container/string object

        :param data: JsON data or UTF-8 string bytes
        :return: Map, List, or String
        """
        raise NotImplemented


"""
    Interfaces
"""


def base64_encode(data: bytes) -> str:
    return Base64.encode(data)


def base64_decode(string: str) -> Optional[bytes]:
    return Base64.decode(string)


def base58_encode(data: bytes) -> str:
    return Base58.encode(data)


def base58_decode(string: str) -> Optional[bytes]:
    return Base58.decode(string)


def hex_encode(data: bytes) -> str:
    return Hex.encode(data)


def hex_decode(string: str) -> Optional[bytes]:
    return Hex.decode(string)


def json_encode(o: Union[dict, list]) -> bytes:
    return JSON.encode(o=o)


def json_decode(data: bytes) -> Union[dict, list, None]:
    return JSON.decode(data=data)


def utf8_encode(string: str) -> bytes:
    return UTF8.encode(string=string)


def utf8_decode(data: bytes) -> Optional[str]:
    return UTF8.decode(data=data)


"""
    Implementations
    ~~~~~~~~~~~~~~~
    
    Data coders
"""


class B64(DataCoder):

    def encode(self, data: bytes) -> str:
        """ BASE-64 Encode """
        return base64.b64encode(data).decode('utf-8')

    def decode(self, string: str) -> Optional[bytes]:
        """ BASE-64 Decode """
        return base64.b64decode(string)


class H(DataCoder):

    def encode(self, data: bytes) -> str:
        """ HEX Encode """
        # return binascii.b2a_hex(data).decode('utf-8')
        return data.hex()

    def decode(self, string: str) -> Optional[bytes]:
        """ HEX Decode """
        # return binascii.a2b_hex(string)
        return bytes.fromhex(string)


class Base64:
    coder: DataCoder = B64()

    @staticmethod
    def encode(data: bytes) -> str:
        assert Base64.coder is not None, 'Base64 coder not set yet'
        return Base64.coder.encode(data=data)

    @staticmethod
    def decode(string: str) -> Optional[bytes]:
        assert Base64.coder is not None, 'Base64 coder not set yet'
        return Base64.coder.decode(string=string)


class Base58:
    coder: DataCoder = None

    @staticmethod
    def encode(data: bytes) -> str:
        assert Base58.coder is not None, 'Base58 coder not set yet'
        return Base58.coder.encode(data=data)

    @staticmethod
    def decode(string: str) -> Optional[bytes]:
        assert Base58.coder is not None, 'Base58 coder not set yet'
        return Base58.coder.decode(string=string)


class Hex:
    coder: DataCoder = H()

    @staticmethod
    def encode(data: bytes) -> str:
        assert Hex.coder is not None, 'Hex coder not set yet'
        return Hex.coder.encode(data=data)

    @staticmethod
    def decode(string: str) -> Optional[bytes]:
        assert Hex.coder is not None, 'Hex coder not set yet'
        return Hex.coder.decode(string=string)


"""
    Implementations
    ~~~~~~~~~~~~~~~

    Data parsers
"""


class J(DataParser):

    def encode(self, o: Union[dict, list]) -> bytes:
        """ JsON encode """
        return bytes(json.dumps(o), encoding='utf-8')

    def decode(self, data: bytes) -> Union[dict, list, None]:
        """ JsON decode """
        return json.loads(data)


class U(DataParser):

    def encode(self, o: str) -> bytes:
        """ UTF-8 encode """
        return o.encode('utf-8')

    def decode(self, data: bytes) -> Optional[str]:
        """ UTF-8 decode """
        return data.decode('utf-8')


class JSON:
    parser: DataParser = J()

    @staticmethod
    def encode(o: Union[dict, list]) -> bytes:
        assert JSON.parser is not None, 'JSON parser not set yet'
        return JSON.parser.encode(o=o)

    @staticmethod
    def decode(data: bytes) -> Union[dict, list, None]:
        assert JSON.parser is not None, 'JSON parser not set yet'
        return JSON.parser.decode(data=data)


class UTF8:
    parser: DataParser = U()

    @staticmethod
    def encode(string: str) -> bytes:
        assert UTF8.parser is not None, 'UTF8 parser not set yet'
        return UTF8.parser.encode(o=string)

    @staticmethod
    def decode(data: bytes) -> Optional[str]:
        assert UTF8.parser is not None, 'UTF8 parser not set yet'
        return UTF8.parser.decode(data=data)



# -*- coding: utf-8 -*-
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import copy
from abc import abstractmethod
from typing import Any, Optional, Union, Tuple, Iterator
from typing import Mapping, MutableMapping, ItemsView, KeysView, ValuesView

from .string import String


class Wrapper:
    """
        Wrapper for dict, list, str
    """

    @classmethod
    def unwrap(cls, o: Any, circularly: bool = False) -> Any:
        # unwrap string
        if isinstance(o, String):
            return o.string
        if isinstance(o, str):
            return o
        # unwrap container
        if circularly:
            # unwrap map circularly
            if isinstance(o, Map):
                return Map.unwrap(o.dictionary, circularly=True)
            if isinstance(o, dict):
                return Map.unwrap(o, circularly=True)
            # unwrap list circularly
            if isinstance(o, Array):
                return Array.unwrap(o.array, circularly=True)
            if isinstance(o, list):
                return Array.unwrap(o, circularly=True)
        else:
            # unwrap map
            if isinstance(o, Map):
                return o.dictionary
            # unwrap list
            if isinstance(o, Array):
                return o.array
        # others
        return o


class Array:
    """
        List: Any
    """

    @classmethod
    def unwrap(cls, a: list, circularly: bool = False) -> list:
        # unwrap list container
        if isinstance(a, Array):
            a = a.array
        if not circularly:
            return a
        # unwrap items circularly
        t = []
        for o in a:
            item = Wrapper.unwrap(o, circularly=True)
            t.append(item)
        return t

    @property
    @abstractmethod
    def array(self) -> list:
        raise NotImplemented

    @property
    @abstractmethod
    def copy_array(self, deep_copy: bool = False) -> list:
        raise NotImplemented


class Map(MutableMapping):
    """
        Mapping: str -> Any
    """

    @classmethod
    def unwrap(cls, d: dict, circularly: bool = False) -> dict:
        # unwrap map container
        if isinstance(d, Map):
            d = d.dictionary
        if not circularly:
            return d
        # unwrap keys, values circularly
        t = {}
        for k in d:
            v = d[k]
            key = String.unwrap(k)
            value = Wrapper.unwrap(v, circularly=True)
            t[key] = value
        return t

    @property
    @abstractmethod
    def dictionary(self) -> dict:
        raise NotImplemented

    @abstractmethod
    def copy_dictionary(self, deep_copy: bool = False) -> dict:
        raise NotImplemented


class Dictionary(Map):
    """
        A container sharing the same inner dictionary
    """

    # def __new__(cls, dictionary: Optional[dict]=None):
    #     return super().__new__(cls, dictionary=dictionary)

    def __init__(self, dictionary: Optional[dict] = None):
        super().__init__()
        if dictionary is None:
            self.__dictionary = {}
        elif isinstance(dictionary, Dictionary):
            self.__dictionary = dictionary.dictionary
        else:
            assert isinstance(dictionary, dict), 'dictionary error: %s' % dictionary
            self.__dictionary = dictionary

    @property
    def dictionary(self) -> dict:
        return self.__dictionary

    def copy_dictionary(self, deep_copy: bool = False) -> dict:
        if deep_copy:
            copy.deepcopy(self.__dictionary)
        else:
            return self.__dictionary.copy()

    def clear(self):
        """ D.clear() -> None.  Remove all items from D. """
        self.__dictionary.clear()

    def copy(self):
        """ D.copy() -> a shallow copy of D """
        dictionary = self.__dictionary.copy()
        return Dictionary(dictionary=dictionary)

    @staticmethod
    def fromkeys(*args, **kwargs):
        """ Create a new dictionary with keys from iterable and values set to value. """
        dictionary = dict.fromkeys(*args, **kwargs)
        return Dictionary(dictionary=dictionary)

    def get(self, k: str, default: Optional[Any] = None) -> Optional[Any]:
        """ Return the value for key if key is in the dictionary, else default. """
        return self.__dictionary.get(k, default)

    def items(self) -> ItemsView[str, Any]:
        """ D.items() -> a set-like object providing a view on D's items """
        return self.__dictionary.items()

    def keys(self) -> KeysView[str]:
        """ D.keys() -> a set-like object providing a view on D's keys """
        return self.__dictionary.keys()

    def pop(self, k: str, default: Optional[Any] = None) -> Optional[Any]:
        """
        D.pop(k[,d]) -> v, remove specified key and return the corresponding value.
        If key is not found, d is returned if given, otherwise KeyError is raised
        """
        return self.__dictionary.pop(k, default)

    def popitem(self) -> Tuple[str, Any]:
        """
        D.popitem() -> (k, v), remove and return some (key, value) pair as a
        2-tuple; but raise KeyError if D is empty.
        """
        return self.__dictionary.popitem()

    def setdefault(self, k: str, default: Optional[Any] = None) -> Any:
        """
        Insert key with a value of default if key is not in the dictionary.

        Return the value for key if key is in the dictionary, else default.
        """
        self.__dictionary.setdefault(k, default)

    def update(self, __m: Mapping[str, Any], **kwargs: Any):
        self.__dictionary.update(__m)

    def values(self) -> ValuesView[Any]:
        """ D.values() -> an object providing a view on D's values """
        return self.__dictionary.values()

    def __contains__(self, o) -> bool:
        """ True if the dictionary has the specified key, else False. """
        return self.__dictionary.__contains__(o)

    def __delitem__(self, v: str):
        """ Delete self[key]. """
        self.__dictionary.__delitem__(v)

    def __eq__(self, o: Union[dict, Map]) -> bool:
        """ Return self==value. """
        if self is o:
            return True
        if isinstance(o, Map):
            o = o.dictionary
        return self.__dictionary.__eq__(o)

    # def __getattribute__(self, name: str) -> Any:
    #     """ Return getattr(self, name). """
    #     if isinstance(name, String):
    #         name = name.string
    #     return self.__dictionary.__getattribute__(name=name)

    def __getitem__(self, k: str) -> Any:
        """ x.__getitem__(y) <==> x[y] """
        return self.__dictionary.__getitem__(k)

    def __ge__(self, other) -> bool:
        """ Return self>=value. """
        pass

    def __gt__(self, other) -> bool:
        """ Return self>value. """
        pass

    def __iter__(self) -> Iterator[str]:
        """ Implement iter(self). """
        return self.__dictionary.__iter__()

    def __len__(self) -> int:
        """ Return len(self). """
        return self.__dictionary.__len__()

    def __le__(self, other) -> bool:
        """ Return self<=value. """
        pass

    def __lt__(self, other) -> bool:
        """ Return self<value. """
        pass

    def __ne__(self, o: Union[dict, Map]) -> bool:
        """ Return self!=value. """
        if isinstance(o, Map):
            if self is o:
                return False
            o = o.dictionary
        return self.__dictionary.__ne__(o)

    def __repr__(self) -> str:
        """ Return repr(self). """
        return self.__dictionary.__repr__()

    def __setitem__(self, k: str, v: Optional[Any]):
        """ Set self[key] to value. """
        self.__dictionary.__setitem__(k, v)

    def __sizeof__(self) -> int:
        """ D.__sizeof__() -> size of D in memory, in bytes """
        return self.__dictionary.__sizeof__()

    __hash__ = None



# -*- coding: utf-8 -*-
#
#   Ming-Ke-Ming : Decentralized User Identity Authentication
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import abstractmethod
from typing import Optional, Dict

from .crypto import String
from .types import NetworkType, network_is_user, network_is_group


class Address:
    """This class is used to build address for ID

        Address for MKM ID
        ~~~~~~~~~~~~~~~~~~

        properties:
            network - address type
    """

    @property
    @abstractmethod
    def network(self) -> int:
        """
        Get network type of address

        :return: integer as NetworkType
        """
        raise NotImplemented

    @property
    def is_broadcast(self) -> bool:
        # return isinstance(self, BroadcastAddress)
        return False

    @property
    def is_user(self) -> bool:
        return network_is_user(network=self.network)

    @property
    def is_group(self) -> bool:
        return network_is_group(network=self.network)

    #
    #   Address factory
    #
    class Factory:

        @abstractmethod
        def parse_address(self, address: str):  # -> Optional[Address]:
            """
            Parse string object to address

            :param address: address string
            :return: Address
            """
            raise NotImplemented

    __factory = None

    @classmethod
    def register(cls, factory: Factory):
        cls.__factory = factory

    @classmethod
    def factory(cls) -> Factory:
        return cls.__factory

    @classmethod
    def parse(cls, address: str):  # -> Address:
        if address is None:
            return None
        elif isinstance(address, cls):
            return address
        elif isinstance(address, String):
            address = address.string
        factory = cls.factory()
        assert factory is not None, 'address factory not ready'
        return factory.parse_address(address=address)


"""
    Address for broadcast
    ~~~~~~~~~~~~~~~~~~~~~
"""


class BroadcastAddress(String, Address):

    def __init__(self, address: str, network: NetworkType):
        super().__init__(string=address)
        self.__network = network.value

    @property
    def network(self) -> int:
        return self.__network

    @property
    def is_broadcast(self) -> bool:
        return True


ANYWHERE = BroadcastAddress(address='anywhere', network=NetworkType.MAIN)
EVERYWHERE = BroadcastAddress(address='everywhere', network=NetworkType.GROUP)


"""
    Address Factory
    ~~~~~~~~~~~~~~~
"""


class AddressFactory(Address.Factory):

    def __init__(self):
        super().__init__()
        # cache broadcast addresses
        self.__addresses: Dict[str, Address] = {
            str(ANYWHERE): ANYWHERE,
            str(EVERYWHERE): EVERYWHERE,
        }

    def parse_address(self, address: str) -> Optional[Address]:
        add = self.__addresses.get(address)
        if add is None:
            add = self.create_address(address=address)
            if add is not None:
                self.__addresses[address] = add
        return add

    @abstractmethod
    def create_address(self, address: str) -> Optional[Address]:
        """ override for creating address from string """
        raise NotImplemented



# -*- coding: utf-8 -*-
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABC, abstractmethod

from .cryptography import CryptographyKey, promise


class AsymmetricKey(CryptographyKey, ABC):

    RSA = 'RSA'
    ECC = 'ECC'


class SignKey(AsymmetricKey):

    @abstractmethod
    def sign(self, data: bytes) -> bytes:
        """
        signature = sign(data, SK);

        :param data: message data
        :return:     signature
        """
        raise NotImplemented


class VerifyKey(AsymmetricKey):

    @abstractmethod
    def verify(self, data: bytes, signature: bytes) -> bool:
        """
        OK = verify(data, signature, PK)

        :param data:      message data
        :param signature: signature of message data
        :return:          True on signature matched
        """
        raise NotImplemented

    def match(self, key: SignKey):
        """
        OK = verify(data, sign(data, SK), PK)

        :param key: private key
        :return:    True on signature matched
        """
        return asymmetric_keys_match(sign_key=key, verify_key=self)


def asymmetric_keys_match(sign_key: SignKey, verify_key: VerifyKey) -> bool:
    # try to verify with signature
    signature = sign_key.sign(data=promise)
    return verify_key.verify(data=promise, signature=signature)



# -*- coding: utf-8 -*-
#
#   Ming-Ke-Ming : Decentralized User Identity Authentication
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import abstractmethod
from typing import Optional, Union

from .crypto import base64_encode, base64_decode, utf8_encode
from .crypto import Map, Dictionary
from .crypto import VerifyKey, SignKey, PublicKey

from .types import NetworkType, MetaType, meta_has_seed
from .address import Address
from .identifier import ID


class Meta(Map):
    """This class is used to generate entity ID

        User/Group Meta data
        ~~~~~~~~~~~~~~~~~~~~

        data format: {
            version: 1,          // meta version
            seed: "moKy",        // user/group name
            key: "{public key}", // PK = secp256k1(SK);
            fingerprint: "..."   // CT = sign(seed, SK);
        }

        algorithm:
            fingerprint = sign(seed, SK);
    """

    @property
    @abstractmethod
    def type(self) -> int:
        """
        Meta algorithm version

            0x01 - username@address
            0x02 - btc_address
            0x03 - username@btc_address
            0x04 - eth_address
            0x05 - username@eth_address
            ...
        """
        raise NotImplemented

    @property
    @abstractmethod
    def key(self) -> VerifyKey:
        """
        Public key (used for signature)

            RSA
            ECC

        :return: public key
        """
        raise NotImplemented

    @property
    @abstractmethod
    def seed(self) -> Optional[str]:
        """
        Seed to generate fingerprint

        :return: ID.name
        """
        raise NotImplemented

    @property
    @abstractmethod
    def fingerprint(self) -> Optional[bytes]:
        """
        Fingerprint to verify ID and public key

            Build: fingerprint = sign(seed, privateKey)
            Check: verify(seed, fingerprint, publicKey)

        :return: signature
        """
        raise NotImplemented

    @property
    @abstractmethod
    def valid(self) -> bool:
        """
        Check meta valid
        (must call this when received a new meta from network)

        :return: True on valid
        """
        raise NotImplemented

    @abstractmethod
    def generate_identifier(self, network: Union[NetworkType, int], terminal: Optional[str] = None) -> Optional[ID]:
        """
        Generate ID with terminal

        :param network:  ID.type
        :param terminal: ID.terminal
        :return: ID
        """
        raise NotImplemented

    @abstractmethod
    def match_identifier(self, identifier: ID) -> bool:
        """
        Check whether meta match with entity ID
        (must call this when received a new meta from network)

        :param identifier: entity ID
        :return: True on matched
        """
        raise NotImplemented

    @abstractmethod
    def match_key(self, key: VerifyKey) -> bool:
        """
        Check whether meta match with public key

        :param key: public key
        :return: True on matched
        """
        raise NotImplemented

    #
    #   Meta factory
    #
    class Factory:

        @abstractmethod
        def create_meta(self, key: VerifyKey, seed: Optional[str] = None,
                        fingerprint: Union[bytes, str, None] = None):  # -> Meta:
            """
            Create meta

            :param key:         public key
            :param seed:        ID.name
            :param fingerprint: sKey.sign(seed)
            :return: Meta
            """
            raise NotImplemented

        @abstractmethod
        def generate_meta(self, key: SignKey, seed: Optional[str] = None):  # -> Meta:
            """
            Generate meta

            :param key:  private key
            :param seed: ID.name
            :return: Meta
            """
            raise NotImplemented

        @abstractmethod
        def parse_meta(self, meta: dict):  # -> Optional[Meta]:
            """
            Parse map object to meta

            :param meta: meta info
            :return: Meta
            """
            raise NotImplemented

    __factories = {}

    @classmethod
    def register(cls, version: Union[MetaType, int], factory: Factory):
        if isinstance(version, MetaType):
            version = version.value
        cls.__factories[version] = factory

    @classmethod
    def factory(cls, version: Union[MetaType, int]) -> Optional[Factory]:
        if isinstance(version, MetaType):
            version = version.value
        return cls.__factories.get(version)

    @classmethod
    def create(cls, version: Union[MetaType, int], key: VerifyKey,
               seed: Optional[str] = None, fingerprint: Union[bytes, str, None] = None):  # -> Optional[Meta]:
        factory = cls.factory(version=version)
        assert factory is not None, 'meta type not support: %d' % version
        return factory.create_meta(key=key, seed=seed, fingerprint=fingerprint)

    @classmethod
    def generate(cls, version: Union[MetaType, int], key: SignKey, seed: Optional[str] = None):  # -> Optional[Meta]:
        factory = cls.factory(version=version)
        assert factory is not None, 'meta type not support: %d' % version
        return factory.generate_meta(key=key, seed=seed)

    @classmethod
    def parse(cls, meta: dict):  # -> Optional[Meta]:
        if meta is None:
            return None
        elif isinstance(meta, cls):
            return meta
        elif isinstance(meta, Map):
            meta = meta.dictionary
        version = meta_type(meta=meta)
        factory = cls.factory(version=version)
        if factory is None:
            factory = cls.factory(version=0)  # unknown
            assert factory is not None, 'cannot parse meta: %s' % meta
        return factory.parse_meta(meta=meta)


"""
    Implements
    ~~~~~~~~~~
"""


def meta_type(meta: dict) -> int:
    version = meta.get('type')
    if version is None:
        version = meta.get('version')
    return int(version)


def meta_key(meta: dict) -> VerifyKey:
    key = meta.get('key')
    return PublicKey.parse(key=key)


def meta_seed(meta: dict) -> Optional[str]:
    return meta.get('seed')


def meta_fingerprint(meta: dict) -> Optional[bytes]:
    fingerprint = meta.get('fingerprint')
    if fingerprint is not None:
        return base64_decode(string=fingerprint)


class BaseMeta(Dictionary, Meta):

    def __init__(self, meta: Optional[dict] = None,
                 version: Union[MetaType, int] = 0, key: Optional[VerifyKey] = None,
                 seed: Optional[str] = None, fingerprint: Union[bytes, str, None] = None):
        super().__init__(dictionary=meta)
        # pre-process
        if isinstance(version, MetaType):
            version = version.value
        if fingerprint is None:
            base64 = None
        elif isinstance(fingerprint, bytes):
            base64 = base64_encode(data=fingerprint)
        else:
            assert isinstance(fingerprint, str), 'meta.fingerprint error: %s' % fingerprint
            base64 = fingerprint
            fingerprint = base64_decode(string=base64)
        # set values
        self.__type = version
        self.__key = key
        self.__seed = seed
        self.__fingerprint = fingerprint
        self.__status = 0  # 1 for valid, -1 for invalid
        # set values to inner dictionary
        if version > 0:
            self['version'] = version
        if key is not None:
            self['key'] = key.dictionary
        if seed is not None:
            self['seed'] = seed
        if fingerprint is not None:
            self['fingerprint'] = base64

    @property
    def has_seed(self) -> bool:
        return meta_has_seed(version=self.type)

    @property
    def type(self) -> int:
        if self.__type == 0:
            self.__type = meta_type(meta=self.dictionary)
        return self.__type

    @property
    def key(self) -> VerifyKey:
        if self.__key is None:
            self.__key = meta_key(meta=self.dictionary)
        return self.__key

    @property
    def seed(self) -> Optional[str]:
        if self.__seed is None and self.has_seed:
            self.__seed = meta_seed(meta=self.dictionary)
        return self.__seed

    @property
    def fingerprint(self) -> Optional[bytes]:
        if self.__fingerprint is None and self.has_seed:
            self.__fingerprint = meta_fingerprint(meta=self.dictionary)
        return self.__fingerprint

    @property
    def valid(self) -> bool:
        if self.__status == 0:
            key = self.key
            if key is None:
                # meta.key should not be empty
                self.__status = -1
            elif self.has_seed:
                seed = self.seed
                fingerprint = self.fingerprint
                if seed is None or fingerprint is None:
                    # seed and fingerprint should not be empty
                    self.__status = -1
                elif key.verify(data=utf8_encode(string=seed), signature=fingerprint):
                    # fingerprint matched
                    self.__status = 1
                else:
                    # fingerprint not matched
                    self.__status = -1
            else:
                # BTC, ETH, ...
                self.__status = 1
        return self.__status == 1

    @abstractmethod
    def generate_address(self, network: Union[NetworkType, int]) -> Address:
        """
        Generate address

        :param network: ID.type
        :return: Address
        """
        raise NotImplemented

    def generate_identifier(self, network: Union[NetworkType, int], terminal: Optional[str] = None) -> Optional[ID]:
        address = self.generate_address(network=network)
        if address is not None:
            return ID.create(address=address, name=self.seed, terminal=terminal)

    def match_identifier(self, identifier: ID) -> bool:
        if self.valid and self.seed == identifier.name:
            return self.generate_address(network=identifier.type) == identifier.address

    def match_key(self, key: VerifyKey) -> bool:
        if not self.valid:
            return False
        if self.key == key:
            return True
        if self.has_seed:
            # check whether keys equal by verifying signature
            seed = self.seed
            fingerprint = self.fingerprint
            return key.verify(data=utf8_encode(string=seed), signature=fingerprint)
        else:
            # NOTICE: ID with BTC/ETH address has no username, so
            #         just compare the key.data to check matching
            return False



# -*- coding: utf-8 -*-
#
#   Ming-Ke-Ming : Decentralized User Identity Authentication
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import abstractmethod
from typing import Optional, Union, Any

from .crypto import Map
from .crypto import VerifyKey, SignKey

from .identifier import ID


class TAI:
    """
        The Additional Information
        ~~~~~~~~~~~~~~~~~~~~~~~~~~

        'Meta' is the information for entity which never changed,
            which contains the key for verify signature;
        'TAI' is the variable part,
            which could contain a public key for asymmetric encryption.
    """

    @property
    @abstractmethod
    def valid(self) -> bool:
        """
        Check if signature matched

        :return: True on matched
        """
        raise NotImplemented

    #
    #  signature
    #

    @abstractmethod
    def verify(self, public_key: VerifyKey) -> bool:
        """
        Verify 'data' and 'signature' with public key

        :param public_key: public key in meta.key
        :return: True on signature matched
        """
        raise NotImplemented

    @abstractmethod
    def sign(self, private_key: SignKey) -> bytes:
        """
        Encode properties to 'data' and sign it to 'signature'

        :param private_key: private key match meta.key
        :return: signature
        """
        raise NotImplemented

    #
    #  properties
    #

    @property
    @abstractmethod
    def properties(self) -> Optional[dict]:
        """
        Get all properties when valid

        :return: inner dictionary
        """
        raise NotImplemented

    @abstractmethod
    def get_property(self, key: str) -> Optional[Any]:
        """
        Get property value with key

        :param key: property key
        :return: property value
        """
        raise NotImplemented

    @abstractmethod
    def set_property(self, key: str, value: Any = None):
        """
        Update property with key and data
        (this will clear 'data' and 'signature')

        :param key:   property key
        :param value: property value
        """
        raise NotImplemented


class Document(TAI, Map):

    #
    #  Document types
    #
    VISA = 'visa'          # for login/communication
    PROFILE = 'profile'    # for user info
    BULLETIN = 'bulletin'  # for group info

    @property
    @abstractmethod
    def type(self) -> str:
        """
        Get document type

        :return: doc type
        """
        raise NotImplemented

    @property
    @abstractmethod
    def identifier(self) -> ID:
        """
        Get entity ID

        :return: Entity ID
        """
        raise NotImplemented

    @property
    @abstractmethod
    def time(self) -> int:
        """
        Get sign time

        :return: timestamp
        """
        raise NotImplemented

    #
    #  properties getter/setter
    #
    @property
    def name(self) -> str:
        """
        Get entity name

        :return: name string
        """
        raise NotImplemented

    @name.setter
    def name(self, string: str):
        """
        Set entity name

        :param string: name string
        :return:
        """
        raise NotImplemented

    #
    #   Document factory
    #
    class Factory:

        @abstractmethod
        def create_document(self, identifier: ID,
                            data: Union[bytes, str, None] = None,
                            signature: Union[bytes, str, None] = None):  # -> Document:
            """
            1. Create a new empty document with entity ID

            2. Create document with data & signature loaded from local storage

            :param identifier: entity ID
            :param data:       document data
            :param signature:  document signature
            :return: Document
            """
            raise NotImplemented

        @abstractmethod
        def parse_document(self, document: dict):  # -> Optional[Document]:
            """
            Parse map object to entity document

            :param document:
            :return:
            """
            raise NotImplemented

    __factories = {}

    @classmethod
    def register(cls, doc_type: str, factory: Factory):
        cls.__factories[doc_type] = factory

    @classmethod
    def factory(cls, doc_type: str) -> Optional[Factory]:
        return cls.__factories.get(doc_type)

    @classmethod
    def create(cls, doc_type: str, identifier: ID, data: Union[bytes, str] = None, signature: Union[bytes, str] = None):
        factory = cls.factory(doc_type=doc_type)
        assert factory is not None, 'doc_type not support: %s' % doc_type
        return factory.create_document(identifier=identifier, data=data, signature=signature)

    @classmethod
    def parse(cls, document: dict):
        if document is None:
            return None
        elif isinstance(document, cls):
            return document
        elif isinstance(document, Map):
            document = document.dictionary
        doc_type = document_type(document=document)
        factory = cls.factory(doc_type=doc_type)
        if factory is None:
            factory = cls.factory(doc_type='*')  # unknown
            assert factory is not None, 'cannot parse document: %s' % document
        return factory.parse_document(document=document)


def document_type(document: dict) -> str:
    return document.get('type')



# -*- coding: utf-8 -*-
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Data Digest
    ~~~~~~~~~~~

    MD5, SHA-256, RipeMD-160
"""

import hashlib
from abc import abstractmethod


class DataDigester:

    @abstractmethod
    def digest(self, data: bytes) -> bytes:
        raise NotImplemented


"""
    Interfaces
"""


def md5(data: bytes) -> bytes:
    return MD5.digest(data=data)


def sha1(data: bytes) -> bytes:
    return SHA1.digest(data=data)


def sha256(data: bytes) -> bytes:
    return SHA256.digest(data=data)


def keccak256(data: bytes) -> bytes:
    return KECCAK256.digest(data=data)


def ripemd160(data: bytes) -> bytes:
    return RIPEMD160.digest(data=data)


"""
    Implementations
"""


class M5(DataDigester):

    def digest(self, data: bytes) -> bytes:
        """ MD5 digest """
        hash_obj = hashlib.md5()
        hash_obj.update(data)
        return hash_obj.digest()


class S1(DataDigester):

    def digest(self, data: bytes) -> bytes:
        """ SHA1 Digest """
        return hashlib.sha1(data).digest()


class S256(DataDigester):

    def digest(self, data: bytes) -> bytes:
        """ SHA-256 """
        return hashlib.sha256(data).digest()


class R160(DataDigester):

    def digest(self, data: bytes) -> bytes:
        """ RIPEMD-160 """
        hash_obj = hashlib.new('ripemd160')
        hash_obj.update(data)
        return hash_obj.digest()


class MD5:
    digester: DataDigester = M5()

    @staticmethod
    def digest(data: bytes) -> bytes:
        assert MD5.digester is not None, 'MD5 coder not set yet'
        return MD5.digester.digest(data=data)


class SHA1:
    digester: DataDigester = S1()

    @staticmethod
    def digest(data: bytes) -> bytes:
        assert SHA1.digester is not None, 'SHA1 coder not set yet'
        return SHA1.digester.digest(data=data)


class SHA256:
    digester: DataDigester = S256()

    @staticmethod
    def digest(data: bytes) -> bytes:
        assert SHA256.digester is not None, 'SHA256 coder not set yet'
        return SHA256.digester.digest(data=data)


class KECCAK256:
    digester: DataDigester = None

    @staticmethod
    def digest(data: bytes) -> bytes:
        assert KECCAK256.digester is not None, 'KECCAK256 coder not set yet'
        return KECCAK256.digester.digest(data=data)


class RIPEMD160:
    digester: DataDigester = R160()

    @staticmethod
    def digest(data: bytes) -> bytes:
        assert RIPEMD160.digester is not None, 'RIPEMD160 coder not set yet'
        return RIPEMD160.digester.digest(data=data)



# -*- coding: utf-8 -*-
#
#   Ming-Ke-Ming : Decentralized User Identity Authentication
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import abstractmethod
from typing import Optional

from .crypto import String
from .address import Address, ANYWHERE, EVERYWHERE


class ID:
    """
        ID for entity (User/Group)
        ~~~~~~~~~~~~~~~~~~~~~~~~~~

        data format: "name@address[/terminal]"

        fields:
            name     - entity name, the seed of fingerprint to generate address
            address  - a string to identify an entity
            terminal - entity login resource(device), OPTIONAL
    """

    # def __eq__(self, other) -> bool:
    #     if other is None:
    #         return False
    #     if not isinstance(other, ID):
    #         if isinstance(other, String):
    #             other = other.string
    #         assert isinstance(other, str), 'ID error: %s' % other
    #         other = ID.parse(identifier=other)
    #         if other is None:
    #             return False
    #     return ID.equals(id1=self, id2=other)
    #
    # @classmethod
    # def equals(cls, id1, id2) -> bool:
    #     # assert isinstance(id1, ID)
    #     # assert isinstance(id2, ID)
    #     if id1 is id2:
    #         # same object
    #         return True
    #     else:
    #         return id1.address == id2.address and id1.name == id2.name

    @property
    @abstractmethod
    def name(self) -> Optional[str]:
        raise NotImplemented

    @property
    @abstractmethod
    def address(self) -> Address:
        raise NotImplemented

    @property
    @abstractmethod
    def terminal(self) -> Optional[str]:
        raise NotImplemented

    @property
    def type(self) -> int:
        """ ID type """
        return self.address.network

    @property
    def is_broadcast(self) -> bool:
        return self.address.is_broadcast

    @property
    def is_user(self) -> bool:
        return self.address.is_user

    @property
    def is_group(self) -> bool:
        return self.address.is_group

    @classmethod
    def convert(cls, members: list) -> list:
        """
        Convert ID list from string array

        :param members: string array
        :return: ID list
        """
        array = []
        for item in members:
            identifier = cls.parse(identifier=item)
            if identifier is None:
                continue
            array.append(identifier)
        return array

    @classmethod
    def revert(cls, members: list) -> list:
        """
        Revert ID list to string array

        :param members: ID list
        :return: string array
        """
        array = []
        for item in members:
            array.append(str(item))
        return array

    #
    #   ID factory
    #
    class Factory:

        @abstractmethod
        def create_identifier(self, address: Address, name: Optional[str] = None, terminal: Optional[str] = None):
            """
            Create ID

            :param address:  ID.address
            :param name:     ID.name
            :param terminal: ID.terminal
            :return: ID
            """
            raise NotImplemented

        @abstractmethod
        def parse_identifier(self, identifier: str):  # -> Optional[ID]:
            """
            Parse string object to ID

            :param identifier: ID string
            :return: ID
            """
            raise NotImplemented

    __factory = None

    @classmethod
    def register(cls, factory: Factory):
        cls.__factory = factory

    @classmethod
    def factory(cls) -> Factory:
        return cls.__factory

    @classmethod
    def create(cls, address: Address, name: Optional[str] = None, terminal: Optional[str] = None):  # -> Optional[ID]:
        factory = cls.factory()
        assert factory is not None, 'ID factory not ready'
        return factory.create_identifier(address=address, name=name, terminal=terminal)

    @classmethod
    def parse(cls, identifier: str):  # -> Optional[ID]:
        if identifier is None:
            return None
        elif isinstance(identifier, cls):
            return identifier
        elif isinstance(identifier, String):
            identifier = identifier.string
        factory = cls.factory()
        assert factory is not None, 'ID factory not ready'
        return factory.parse_identifier(identifier=identifier)


"""
    Implements
    ~~~~~~~~~~
"""


def parse(string: str) -> Optional[ID]:
    # split ID string
    pair = string.split('/', 1)
    # terminal
    if len(pair) == 1:
        # no terminal
        terminal = None
    else:
        # got terminal
        terminal = pair[1]
    # name @ address
    assert len(pair[0]) > 0, 'ID error: %s' % string
    pair = pair[0].split('@', 1)
    if len(pair) == 1:
        # got address without name
        name = None
        address = Address.parse(address=pair[0])
    else:
        # got name & address
        name = pair[0]
        address = Address.parse(address=pair[1])
    if address is not None:
        return Identifier(identifier=string, address=address, name=name, terminal=terminal)


def concat(address: Address, name: Optional[str] = None, terminal: Optional[str] = None) -> str:
    string = str(address)
    if name is not None and len(name) > 0:
        string = name + '@' + string
    if terminal is not None and len(terminal) > 0:
        string = string + '/' + terminal
    return string


class Identifier(String, ID):

    def __init__(self, identifier: str, address: Address, name: Optional[str] = None, terminal: Optional[str] = None):
        super().__init__(string=identifier)
        self.__name = name
        self.__address = address
        self.__terminal = terminal

    @property
    def name(self) -> Optional[str]:
        return self.__name

    @property
    def address(self) -> Address:
        return self.__address

    @property
    def terminal(self) -> Optional[str]:
        return self.__terminal


"""
    ID factory
    ~~~~~~~~~~
"""


class IDFactory(ID.Factory):

    def __init__(self):
        super().__init__()
        self.__ids = {}

    def create_identifier(self, address: Address, name: Optional[str] = None, terminal: Optional[str] = None) -> ID:
        identifier = concat(address=address, name=name, terminal=terminal)
        _id = self.__ids.get(identifier)
        if _id is None:
            _id = Identifier(identifier=identifier, address=address, name=name, terminal=terminal)
            self.__ids[identifier] = _id
        return _id

    def parse_identifier(self, identifier: str) -> Optional[ID]:
        _id = self.__ids.get(identifier)
        if _id is None:
            _id = parse(string=identifier)
            if _id is not None:
                self.__ids[identifier] = _id
        return _id


"""
    ID for Broadcast
    ~~~~~~~~~~~~~~~~
"""

# register ID factory
ID.register(factory=IDFactory())

ANYONE = ID.create(name='anyone', address=ANYWHERE)
EVERYONE = ID.create(name='everyone', address=EVERYWHERE)

# DIM Founder
FOUNDER = ID.create(name='moky', address=ANYWHERE)



# -*- coding: utf-8 -*-
#
#   Ming-Ke-Ming : Decentralized User Identity Authentication
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from enum import IntEnum


class NetworkType(IntEnum):
    """
        @enum MKMNetworkID

        @abstract A network type to indicate what kind the entity is.

        @discussion An address can identify a person, a group of people,
            a team, even a thing.

            MKMNetwork_Main indicates this entity is a person's account.
            An account should have a public key, which proved by meta data.

            MKMNetwork_Group indicates this entity is a group of people,
            which should have a founder (also the owner), and some members.

            MKMNetwork_Moments indicates a special personal social network,
            where the owner can share information and interact with its friends.
            The owner is the king here, it can do anything and no one can stop it.

            MKMNetwork_Polylogue indicates a virtual (temporary) social network.
            It's created to talk with multi-people (but not too much, e.g. < 100).
            Any member can invite people in, but only the founder can expel member.

            MKMNetwork_Chatroom indicates a massive (persistent) social network.
            It's usually more than 100 people in it, so we need administrators
            to help the owner to manage the group.

            MKMNetwork_SocialEntity indicates this entity is a social entity.

            MKMNetwork_Organization indicates an independent organization.

            MKMNetwork_Company indicates this entity is a company.

            MKMNetwork_School indicates this entity is a school.

            MKMNetwork_Government indicates this entity is a government department.

            MKMNetwork_Department indicates this entity is a department.

            MKMNetwork_Thing this is reserved for IoT (Internet of Things).

        Bits:
            0000 0001 - this entity's branch is independent (clear division).
            0000 0010 - this entity can contains other group (big organization).
            0000 0100 - this entity is top organization.
            0000 1000 - (MAIN) this entity acts like a human.

            0001 0000 - this entity contains members (Group)
            0010 0000 - this entity needs other administrators (big organization)
            0100 0000 - this is an entity in reality.
            1000 0000 - (IoT) this entity is a 'Thing'.

            (All above are just some advices to help choosing numbers :P)
    """
    ################################
    #  BTC Network
    ################################
    BTC_MAIN = 0x00         # 0000 0000 (BitCoin Address)
    # BTC_TEST = 0x6f       # 0110 1111 (BitCoin Test)

    ################################
    #  Person Account
    ################################
    MAIN = 0x08             # 0000 1000 (Person)

    ################################
    #  Virtual Groups
    ################################
    GROUP = 0x10            # 0001 0000 (Multi-Persons)
    # MOMENTS = 0x18        # 0001 1000 (Twitter)
    POLYLOGUE = 0x10        # 0001 0000 (Multi-Persons Chat, N < 100)
    CHATROOM = 0x30         # 0011 0000 (Multi-Persons Chat, N >= 100)

    ################################
    #  Social Entities in Reality
    ################################
    # SOCIAL_ENTITY = 0x50  # 0101 0000
    # ORGANIZATION = 0x74   # 0111 0100
    # COMPANY = 0x76        # 0111 0110
    # SCHOOL = 0x77         # 0111 0111
    # GOVERNMENT = 0x73     # 0111 0011
    # DEPARTMENT = 0x52     # 0101 0010

    ################################
    #  Network
    ################################
    PROVIDER = 0x76         # 0111 0110 (Service Provider)
    STATION = 0x88          # 1000 1000 (Server Node)

    ################################
    #  Internet of Things
    ################################
    THING = 0x80            # 1000 0000 (IoT)
    ROBOT = 0xC8            # 1100 1000


def network_is_user(network: int) -> bool:
    return (network & NetworkType.MAIN) == NetworkType.MAIN or network == NetworkType.BTC_MAIN


def network_is_group(network: int) -> bool:
    return (network & NetworkType.GROUP) == NetworkType.GROUP


class MetaType(IntEnum):
    """
        @enum MKMMetaVersion

        @abstract Defined for meta data structure to generate identifier.

        @discussion Generate & check ID/Address

            MKMMetaVersion_MKM give a seed string first, and sign this seed to get
            fingerprint; after that, use the fingerprint to generate address.
            This will get a firmly relationship between (username, address & key).

            MKMMetaVersion_BTC use the key data to generate address directly.
            This can build a BTC address for the entity ID without username.

            MKMMetaVersion_ExBTC use the key data to generate address directly, and
            sign the seed to get fingerprint (just for binding username & key).
            This can build an entity ID with username and BTC address.

        Bits:
            0000 0001 - this meta contains seed as ID.name
            0000 0010 - this meta generate BTC address
            0000 0100 - this meta generate ETH address
            ...
    """
    DEFAULT = 0x01
    MKM = 0x01    # 0000 0001

    BTC = 0x02    # 0000 0010
    ExBTC = 0x03  # 0000 0011

    ETH = 0x04    # 0000 0100
    ExETH = 0x05  # 0000 0101


def meta_has_seed(version: int) -> bool:
    return (version & MetaType.MKM) == MetaType.MKM



# -*- coding: utf-8 -*-
#
#   Ming-Ke-Ming : Decentralized User Identity Authentication
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import time as time_lib
from abc import abstractmethod
from typing import Optional, Union, Any, List

from .crypto import json_encode, json_decode, utf8_encode, utf8_decode, base64_encode, base64_decode
from .crypto import Dictionary
from .crypto import PublicKey, EncryptKey, VerifyKey, SignKey

from .identifier import ID
from .tai import Document, document_type


class Visa(Document):
    """
        User Document
        ~~~~~~~~~~~~~
        This interface is defined for authorizing other apps to login,
        which can generate a temporary asymmetric key pair for messaging.
    """

    @property
    @abstractmethod
    def key(self) -> Union[EncryptKey, VerifyKey, None]:
        """
        Get public key to encrypt message for user

        :return: public key
        """
        raise NotImplemented

    @key.setter
    @abstractmethod
    def key(self, value: Union[EncryptKey, VerifyKey]):
        """
        Set public key for other user to encrypt message

        :param value: public key as visa.key
        """
        raise NotImplemented

    @property
    def avatar(self) -> Optional[str]:
        """
        Get avatar URL

        :return: URL string
        """
        return None

    @avatar.setter
    def avatar(self, url: str):
        """
        Set avatar URL

        :param url: URL string
        """
        pass


class Bulletin(Document):
    """
        Group Document
        ~~~~~~~~~~~~~~
    """

    @property
    @abstractmethod
    def assistants(self) -> Optional[List[ID]]:
        """
        Get group assistants

        :return: bot ID list
        """
        raise NotImplemented

    @assistants.setter
    def assistants(self, bots: List[ID]):
        """
        Set group assistants

        :param bots: bot ID list
        """
        raise NotImplemented


"""
    Implements
    ~~~~~~~~~~
"""


def document_identifier(document: dict) -> ID:
    return ID.parse(identifier=document.get('ID'))


def document_data(document: dict) -> Optional[bytes]:
    utf8 = document.get('data')
    if utf8 is not None:
        return utf8_encode(string=utf8)


def document_signature(document: dict) -> Optional[bytes]:
    base64 = document.get('signature')
    if base64 is not None:
        return base64_decode(string=base64)


class BaseDocument(Dictionary, Document):

    def __init__(self, document: Optional[dict] = None,
                 doc_type: Optional[str] = None, identifier: Optional[ID] = None,
                 data: Union[bytes, str, None] = None, signature: Union[bytes, str, None] = None):
        super().__init__(dictionary=document)
        # pre-process
        if data is None:
            utf8 = None
        elif isinstance(data, bytes):
            utf8 = utf8_decode(data=data)
        else:
            assert isinstance(data, str), 'document data error: %s' % data
            utf8 = data
            data = utf8_encode(string=utf8)
        if signature is None:
            base64 = None
        elif isinstance(signature, bytes):
            base64 = base64_encode(data=signature)
        else:
            assert isinstance(signature, str), 'document signature error: %s' % signature
            base64 = signature
            signature = base64_decode(string=base64)
        # set values
        self.__identifier = identifier
        self.__data = data            # JsON.encode(properties)
        self.__signature = signature  # LocalUser(identifier).sign(data)
        self.__properties = None
        self.__status = 0             # 1 for valid, -1 for invalid
        # set values to inner dictionary
        if identifier is not None:
            self['ID'] = str(identifier)
            if data is None or signature is None:
                """ Create a new empty document with ID and doc type """
                assert doc_type is not None, 'document type empty'
                self.__properties = {
                    'type': doc_type,
                }
            else:
                """ Create entity document with ID, data and signature loaded from local storage """
                self['data'] = utf8
                self['signature'] = base64
                self.__status = 1  # all documents must be verified before saving into local storage

    @property
    def type(self) -> str:
        doc_type = self.get_property(key='type')
        if doc_type is None:
            doc_type = document_type(document=self.dictionary)
        return doc_type

    @property
    def identifier(self) -> ID:
        if self.__identifier is None:
            self.__identifier = document_identifier(document=self.dictionary)
        return self.__identifier

    @property
    def data(self) -> Optional[bytes]:
        """
        Get serialized properties

        :return: JsON string
        """
        if self.__data is None:
            self.__data = document_data(document=self.dictionary)
        return self.__data

    @property
    def signature(self) -> Optional[bytes]:
        """
        Get signature for serialized properties

        :return: signature data
        """
        if self.__signature is None:
            self.__signature = document_signature(document=self.dictionary)
        return self.__signature

    @property
    def valid(self) -> bool:
        return self.__status > 0

    #
    #  signature
    #

    def verify(self, public_key: VerifyKey) -> bool:
        """
        Verify 'data' and 'signature' with public key

        :param public_key: public key in meta.key
        :return: True on signature matched
        """
        if self.__status > 0:
            # already verify OK
            return True
        data = self.data
        signature = self.signature
        if data is None:
            # NOTICE: if data is empty, signature should be empty at the same time
            #         this happen while profile not found
            if signature is None:
                self.__status = 0
            else:
                # data signature error
                self.__status = -1
        elif signature is None:
            # signature error
            self.__status = -1
        elif public_key.verify(data=data, signature=signature):
            # signature matched
            self.__status = 1
        # NOTICE: if status is 0, it doesn't mean the profile is invalid,
        #         try another key
        return self.__status == 1

    def sign(self, private_key: SignKey) -> bytes:
        """
        Encode properties to 'data' and sign it to 'signature'

        :param private_key: private key match meta.key
        :return: signature
        """
        if self.__status > 0:
            # already signed/verified
            return self.__signature
        # update sign time
        self.set_property(key='time', value=int(time_lib.time()))
        # update status
        self.__status = 1
        # sign
        self.__data = json_encode(self.properties)
        self.__signature = private_key.sign(data=self.__data)
        self['data'] = utf8_decode(data=self.__data)  # JsON string
        self['signature'] = base64_encode(data=self.__signature)
        return self.__signature

    #
    #  properties
    #

    @property
    def properties(self) -> Optional[dict]:
        """ Load properties from data """
        if self.__status < 0:
            # invalid
            return None
        if self.__properties is None:
            data = self.data
            if data is None:
                # create new properties
                self.__properties = {}
            else:
                # get properties from data
                self.__properties = json_decode(data=data)
                assert isinstance(self.__properties, dict), 'document data error: %s' % self
        return self.__properties

    def get_property(self, key: str) -> Optional[Any]:
        info = self.properties
        if info is not None:
            return info.get(key)

    def set_property(self, key: str, value: Any = None):
        """ Update profile property with key and value """
        # 1. reset status
        self.__status = 0
        # 2. update property value with name
        info = self.properties
        assert isinstance(info, dict), 'failed to get properties: %s' % self
        if value is None:
            info.pop(key, None)
        else:
            info[key] = value
        # 3. clear data signature after properties changed
        self.pop('data', None)
        self.pop('signature', None)
        self.__data = None
        self.__signature = None

    #
    #  properties getter/setter
    #

    @property
    def time(self) -> int:
        timestamp = self.get_property(key='time')
        if timestamp is None:
            return 0
        else:
            return int(timestamp)

    @property
    def name(self) -> Optional[str]:
        return self.get_property(key='name')

    @name.setter
    def name(self, value: str):
        self.set_property(key='name', value=value)


class BaseVisa(BaseDocument, Visa):

    def __init__(self, document: Optional[dict] = None, identifier: Optional[ID] = None,
                 data: Union[bytes, str, None] = None, signature: Union[bytes, str, None] = None):
        super().__init__(document, Document.VISA, identifier=identifier, data=data, signature=signature)
        self.__key = None

    """
        Public Key for encryption
        ~~~~~~~~~~~~~~~~~~~~~~~~~
        For safety considerations, the profile.key which used to encrypt message data
        should be different with meta.key
    """
    @property
    def key(self) -> Union[EncryptKey, VerifyKey, None]:
        if self.__key is None:
            info = self.get_property(key='key')
            if info is not None:
                self.__key = PublicKey.parse(key=info)
        return self.__key

    @key.setter
    def key(self, value: Union[EncryptKey, VerifyKey]):
        self.set_property(key='key', value=value.dictionary)
        self.__key = value

    """
        Avatar
        ~~~~~~
    """
    @property
    def avatar(self) -> Optional[str]:
        return self.get_property(key='avatar')

    @avatar.setter
    def avatar(self, value: str):
        self.set_property(key='avatar', value=value)


class BaseBulletin(BaseDocument, Bulletin):

    def __init__(self, document: Optional[dict] = None, identifier: Optional[ID] = None,
                 data: Union[bytes, str, None] = None, signature: Union[bytes, str, None] = None):
        super().__init__(document, Document.BULLETIN, identifier=identifier, data=data, signature=signature)
        self.__assistants = None

    @property
    def assistants(self) -> Optional[List[ID]]:
        if self.__assistants is None:
            assistants = self.get_property(key='assistants')
            if assistants is not None:
                self.__assistants = ID.convert(members=assistants)
        return self.__assistants

    @assistants.setter
    def assistants(self, bots: List[ID]):
        self.set_property(key='assistants', value=ID.revert(members=bots))
        self.__assistants = bots



# -*- coding: utf-8 -*-
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from typing import Optional, Union, Any, Tuple, Mapping, List, Iterable, Iterator


class String:
    """
        A container with inner string
    """

    @classmethod
    def unwrap(cls, s: str) -> str:
        # unwrap string container
        if isinstance(s, String):
            return s.string
        # assert isinstance(s, str)
        return s

    # def __new__(cls, string: Optional[str]=None):
    #     return super().__new__(cls, string=string)

    def __init__(self, string: Optional[str] = None):
        super().__init__()
        if string is None:
            self.__string = ''
        elif isinstance(string, String):
            self.__string = string.string
        else:
            assert isinstance(string, str), 'string error: %s' % string
            self.__string = string

    @property
    def string(self) -> str:
        return self.__string

    def capitalize(self):
        """
        Return a capitalized version of the string.

        More specifically, make the first character have upper case and the rest lower
        case.
        """
        string = self.__string.capitalize()
        return String(string=string)

    def casefold(self):
        """ Return a version of the string suitable for caseless comparisons. """
        string = self.__string.casefold()
        return String(string=string)

    def center(self, width: int, fillchar: str = ' '):
        """
        Return a centered string of length width.

        Padding is done using the specified fill character (default is a space).
        """
        if isinstance(fillchar, String):
            fillchar = fillchar.string
        string = self.__string.center(width, fillchar)
        return String(string=string)

    def count(self, x: str, __start: Optional[int] = None, __end: Optional[int] = None) -> int:
        """
        S.count(sub[, start[, end]]) -> int

        Return the number of non-overlapping occurrences of substring sub in
        string S[start:end].  Optional arguments start and end are
        interpreted as in slice notation.
        """
        if isinstance(x, String):
            x = x.string
        return self.__string.count(x, __start, __end)

    def encode(self, encoding: str = 'utf-8', errors: str = 'strict') -> bytes:
        """
        Encode the string using the codec registered for encoding.

          encoding
            The encoding in which to encode the string.
          errors
            The error handling scheme to use for encoding errors.
            The default is 'strict' meaning that encoding errors raise a
            UnicodeEncodeError.  Other possible values are 'ignore', 'replace' and
            'xmlcharrefreplace' as well as any other name registered with
            codecs.register_error that can handle UnicodeEncodeErrors.
        """
        return self.__string.encode(encoding, errors)

    def endswith(self, suffix: str, start: Optional[int] = None, end: Optional[int] = None) -> bool:
        """
        S.endswith(suffix[, start[, end]]) -> bool

        Return True if S ends with the specified suffix, False otherwise.
        With optional start, test S beginning at that position.
        With optional end, stop comparing S at that position.
        suffix can also be a tuple of strings to try.
        """
        if isinstance(suffix, String):
            suffix = suffix.string
        return self.__string.endswith(suffix, start, end)

    def expandtabs(self, tabsize: int = 8):
        """
        Return a copy where all tab characters are expanded using spaces.

        If tabsize is not given, a tab size of 8 characters is assumed.
        """
        string = self.__string.expandtabs(tabsize)
        return String(string=string)

    def find(self, sub: str, __start: Optional[int] = None, __end: Optional[int] = None) -> int:
        """
        S.find(sub[, start[, end]]) -> int

        Return the lowest index in S where substring sub is found,
        such that sub is contained within S[start:end].  Optional
        arguments start and end are interpreted as in slice notation.

        Return -1 on failure.
        """
        if isinstance(sub, String):
            sub = sub.string
        return self.__string.find(sub, __start, __end)

    def format(self, *args, **kwargs):
        """
        S.format(*args, **kwargs) -> str

        Return a formatted version of S, using substitutions from args and kwargs.
        The substitutions are identified by braces ('{' and '}').
        """
        string = self.__string.format(args, kwargs)
        return String(string=string)

    def format_map(self, mapping: Mapping[str, Any]):
        """
        S.format_map(mapping) -> str

        Return a formatted version of S, using substitutions from mapping.
        The substitutions are identified by braces ('{' and '}').
        """
        from .dictionary import Map
        if isinstance(mapping, Map):
            mapping = mapping.dictionary
        string = self.__string.format_map(map=mapping)
        return String(string=string)

    def index(self, sub: str, __start: Optional[int] = None, __end: Optional[int] = None) -> int:
        """
        S.index(sub[, start[, end]]) -> int

        Return the lowest index in S where substring sub is found,
        such that sub is contained within S[start:end].  Optional
        arguments start and end are interpreted as in slice notation.

        Raises ValueError when the substring is not found.
        """
        if isinstance(sub, String):
            sub = sub.string
        return self.__string.index(sub, __start, __end)

    def isalnum(self) -> bool:
        """
        Return True if the string is an alpha-numeric string, False otherwise.

        A string is alpha-numeric if all characters in the string are alpha-numeric and
        there is at least one character in the string.
        """
        return self.__string.isalnum()

    def isalpha(self) -> bool:
        """
        Return True if the string is an alphabetic string, False otherwise.

        A string is alphabetic if all characters in the string are alphabetic and there
        is at least one character in the string.
        """
        return self.__string.isalpha()

    def isascii(self) -> bool:
        """
        Return True if all characters in the string are ASCII, False otherwise.

        ASCII characters have code points in the range U+0000-U+007F.
        Empty string is ASCII too.
        """
        pass

    def isdecimal(self) -> bool:
        """
        Return True if the string is a decimal string, False otherwise.

        A string is a decimal string if all characters in the string are decimal and
        there is at least one character in the string.
        """
        return self.__string.isdecimal()

    def isdigit(self) -> bool:
        """
        Return True if the string is a digit string, False otherwise.

        A string is a digit string if all characters in the string are digits and there
        is at least one character in the string.
        """
        return self.__string.isdigit()

    def isidentifier(self) -> bool:
        """
        Return True if the string is a valid Python identifier, False otherwise.

        Use keyword.iskeyword() to test for reserved identifiers such as "def" and
        "class".
        """
        return self.__string.isidentifier()

    def islower(self) -> bool:
        """
        Return True if the string is a lowercase string, False otherwise.

        A string is lowercase if all cased characters in the string are lowercase and
        there is at least one cased character in the string.
        """
        return self.__string.islower()

    def isnumeric(self) -> bool:
        """
        Return True if the string is a numeric string, False otherwise.

        A string is numeric if all characters in the string are numeric and there is at
        least one character in the string.
        """
        return self.__string.isnumeric()

    def isprintable(self) -> bool:
        """
        Return True if the string is printable, False otherwise.

        A string is printable if all of its characters are considered printable in
        repr() or if it is empty.
        """
        return self.__string.isprintable()

    def isspace(self) -> bool:
        """
        Return True if the string is a whitespace string, False otherwise.

        A string is whitespace if all characters in the string are whitespace and there
        is at least one character in the string.
        """
        return self.__string.isspace()

    def istitle(self) -> bool:
        """
        Return True if the string is a title-cased string, False otherwise.

        In a title-cased string, upper- and title-case characters may only
        follow uncased characters and lowercase characters only cased ones.
        """
        return self.__string.istitle()

    def isupper(self) -> bool:
        """
        Return True if the string is an uppercase string, False otherwise.

        A string is uppercase if all cased characters in the string are uppercase and
        there is at least one cased character in the string.
        """
        return self.__string.isupper()

    def join(self, iterable: Iterable[str]):
        """
        Concatenate any number of strings.

        The string whose method is called is inserted in between each given string.
        The result is returned as a new string.

        Example: '.'.join(['ab', 'pq', 'rs']) -> 'ab.pq.rs'
        """
        string = self.__string.join(iterable)
        return String(string=string)

    def ljust(self, width: int, fillchar: str = ' '):
        """
        Return a left-justified string of length width.

        Padding is done using the specified fill character (default is a space).
        """
        if isinstance(fillchar, String):
            fillchar = fillchar.string
        string = self.__string.ljust(width, fillchar)
        return String(string=string)

    def lower(self):
        """ Return a copy of the string converted to lowercase. """
        string = self.__string.lower()
        return String(string=string)

    def lstrip(self, chars: Optional[str] = None):
        """
        Return a copy of the string with leading whitespace removed.

        If chars is given and not None, remove characters in chars instead.
        """
        if isinstance(chars, String):
            chars = chars.string
        string = self.__string.lstrip(chars)
        return String(string=string)

    def maketrans(self, *args, **kwargs):
        """
        Return a translation table usable for str.translate().

        If there is only one argument, it must be a dictionary mapping Unicode
        ordinals (integers) or characters to Unicode ordinals, strings or None.
        Character keys will be then converted to ordinals.
        If there are two arguments, they must be strings of equal length, and
        in the resulting dictionary, each character in x will be mapped to the
        character at the same position in y. If there is a third argument, it
        must be a string, whose characters will be mapped to None in the result.
        """
        pass

    def partition(self, sep: str) -> Tuple[str, str, str]:
        """
        Partition the string into three parts using the given separator.

        This will search for the separator in the string.  If the separator is found,
        returns a 3-tuple containing the part before the separator, the separator
        itself, and the part after it.

        If the separator is not found, returns a 3-tuple containing the original string
        and two empty strings.
        """
        if isinstance(sep, String):
            sep = sep.string
        return self.__string.partition(sep)

    def replace(self, old: str, new: str, count: int = -1):
        """
        Return a copy with all occurrences of substring old replaced by new.

          count
            Maximum number of occurrences to replace.
            -1 (the default value) means replace all occurrences.

        If the optional argument count is given, only the first count occurrences are
        replaced.
        """
        if isinstance(old, String):
            old = old.string
        if isinstance(new, String):
            new = new.string
        string = self.__string.replace(old, new, count)
        return String(string=string)

    def rfind(self, sub: str, __start: Optional[int] = None, __end: Optional[int] = None) -> int:
        """
        S.rfind(sub[, start[, end]]) -> int

        Return the highest index in S where substring sub is found,
        such that sub is contained within S[start:end].  Optional
        arguments start and end are interpreted as in slice notation.

        Return -1 on failure.
        """
        if isinstance(sub, String):
            sub = sub.string
        return self.__string.rfind(sub, __start, __end)

    def rindex(self, sub: str, __start: Optional[int] = None, __end: Optional[int] = None) -> int:
        """
        S.rindex(sub[, start[, end]]) -> int

        Return the highest index in S where substring sub is found,
        such that sub is contained within S[start:end].  Optional
        arguments start and end are interpreted as in slice notation.

        Raises ValueError when the substring is not found.
        """
        if isinstance(sub, String):
            sub = sub.string
        return self.__string.rindex(sub, __start, __end)

    def rjust(self, width: int, fillchar: str = ' '):
        """
        Return a right-justified string of length width.

        Padding is done using the specified fill character (default is a space).
        """
        if isinstance(fillchar, String):
            fillchar = fillchar.string
        string = self.__string.rjust(width, fillchar)
        return String(string=string)

    def rpartition(self, sep: str) -> Tuple[str, str, str]:
        """
        Partition the string into three parts using the given separator.

        This will search for the separator in the string, starting at the end. If
        the separator is found, returns a 3-tuple containing the part before the
        separator, the separator itself, and the part after it.

        If the separator is not found, returns a 3-tuple containing two empty strings
        and the original string.
        """
        if isinstance(sep, String):
            sep = sep.string
        return self.__string.rpartition(sep)

    def rsplit(self, sep: Optional[str] = None, maxsplit: int = -1) -> List[str]:
        """
        Return a list of the words in the string, using sep as the delimiter string.

          sep
            The delimiter according which to split the string.
            None (the default value) means split according to any whitespace,
            and discard empty strings from the result.
          maxsplit
            Maximum number of splits to do.
            -1 (the default value) means no limit.

        Splits are done starting at the end of the string and working to the front.
        """
        if isinstance(sep, String):
            sep = sep.string
        return self.__string.rsplit(sep, maxsplit)

    def rstrip(self, chars: Optional[str] = None):
        """
        Return a copy of the string with trailing whitespace removed.

        If chars is given and not None, remove characters in chars instead.
        """
        if isinstance(chars, String):
            chars = chars.string
        string = self.__string.rstrip(chars)
        return String(string=string)

    def split(self, sep: Optional[str] = None, maxsplit: int = -1) -> List[str]:
        """
        Return a list of the words in the string, using sep as the delimiter string.

          sep
            The delimiter according which to split the string.
            None (the default value) means split according to any whitespace,
            and discard empty strings from the result.
          maxsplit
            Maximum number of splits to do.
            -1 (the default value) means no limit.
        """
        if isinstance(sep, String):
            sep = sep.string
        return self.__string.split(sep, maxsplit)

    def splitlines(self, keepends: bool = False) -> List[str]:
        """
        Return a list of the lines in the string, breaking at line boundaries.

        Line breaks are not included in the resulting list unless keepends is given and
        true.
        """
        return self.__string.splitlines(keepends)

    def startswith(self, prefix, start: Optional[int] = None, end: Optional[int] = None) -> bool:
        """
        S.startswith(prefix[, start[, end]]) -> bool

        Return True if S starts with the specified prefix, False otherwise.
        With optional start, test S beginning at that position.
        With optional end, stop comparing S at that position.
        prefix can also be a tuple of strings to try.
        """
        if isinstance(prefix, String):
            prefix = prefix.string
        return self.__string.startswith(prefix, start, end)

    def strip(self, chars: Optional[str] = None):
        """
        Return a copy of the string with leading and trailing whitespace remove.

        If chars is given and not None, remove characters in chars instead.
        """
        if isinstance(chars, String):
            chars = chars.string
        string = self.__string.strip(chars)
        return String(string=string)

    def swapcase(self):
        """ Convert uppercase characters to lowercase and lowercase characters to uppercase. """
        string = self.__string.swapcase()
        return String(string=string)

    def title(self):
        """
        Return a version of the string where each word is titlecased.

        More specifically, words start with uppercased characters and all remaining
        cased characters have lower case.
        """
        string = self.__string.title()
        return String(string=string)

    def translate(self):
        """
        Replace each character in the string using the given translation table.

          table
            Translation table, which must be a mapping of Unicode ordinals to
            Unicode ordinals, strings, or None.

        The table must implement lookup/indexing via __getitem__, for instance a
        dictionary or list.  If this operation raises LookupError, the character is
        left untouched.  Characters mapped to None are deleted.
        """
        pass

    def upper(self):
        """ Return a copy of the string converted to uppercase. """
        string = self.__string.upper()
        return String(string=string)

    def zfill(self, width: int):
        """
        Pad a numeric string with zeros on the left, to fill a field of the given width.

        The string is never truncated.
        """
        string = self.__string.zfill(width)
        return String(string=string)

    def __add__(self, s: str):
        """ Return self+value. """
        if isinstance(s, String):
            s = s.string
        string = self.__string.__add__(s)
        return String(string=string)

    def __contains__(self, s: str) -> bool:
        """ Return key in self. """
        if isinstance(s, String):
            s = s.string
        return self.__string.__contains__(s)

    def __eq__(self, x: str) -> bool:
        """ Return self==value. """
        if self is x:
            return True
        if isinstance(x, String):
            x = x.string
        return self.__string.__eq__(x)

    def __format__(self, format_spec: str):
        """ Return a formatted version of the string as described by format_spec. """
        if isinstance(format_spec, String):
            format_spec = format_spec.string
        string = self.__string.__format__(format_spec)
        return String(string=string)

    # def __getattribute__(self, name: str) -> Any:
    #     """ Return getattr(self, name). """
    #     if isinstance(name, String):
    #         name = name.string
    #     return self.__string.__getattribute__(name=name)

    def __getitem__(self, i: Union[int, slice]) -> str:
        """ Return self[key]. """
        return self.__string.__getitem__(i)

    def __getnewargs__(self):
        pass

    def __ge__(self, x: str) -> bool:
        """ Return self>=value. """
        if isinstance(x, String):
            if self is x:
                return True
            x = x.string
        return self.__string.__ge__(x)

    def __gt__(self, x: str) -> bool:
        """ Return self>value. """
        if isinstance(x, String):
            if self is x:
                return False
            x = x.string
        return self.__string.__gt__(x)

    def __hash__(self) -> int:
        """ Return hash(self). """
        return self.__string.__hash__()

    def __iter__(self) -> Iterator[str]:
        """ Implement iter(self). """
        return self.__string.__iter__()

    def __len__(self) -> int:
        """ Return len(self). """
        return self.__string.__len__()

    def __le__(self, x: str) -> bool:
        """ Return self<=value. """
        if isinstance(x, String):
            if self is x:
                return True
            x = x.string
        return self.__string.__le__(x)

    def __lt__(self, x: str) -> bool:
        """ Return self<value. """
        if isinstance(x, String):
            if self is x:
                return False
            x = x.string
        return self.__string.__lt__(x)

    def __mod__(self, value: Any):
        """ Return self%value. """
        string = self.__string.__mod__(value)
        return String(string=string)

    def __mul__(self, n: int):
        """ Return self*value. """
        string = self.__string.__mul__(n)
        return String(string=string)

    def __ne__(self, x: str) -> bool:
        """ Return self!=value. """
        if isinstance(x, String):
            if self is x:
                return False
            x = x.string
        return self.__string.__ne__(x)

    def __repr__(self) -> str:
        """ Return repr(self). """
        return self.__string

    def __rmod__(self, value: Any):
        """ Return value%self. """
        pass

    def __rmul__(self, n: int):
        """ Return value*self. """
        string = self.__string.__rmul__(n)
        return String(string=string)

    def __sizeof__(self) -> int:
        """ Return the size of the string in memory, in bytes. """
        return self.__string.__sizeof__()

    def __str__(self) -> str:
        """ Return str(self). """
        return self.__string



# -*- coding: utf-8 -*-
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABC, abstractmethod
from typing import Optional

from .dictionary import Map
from .cryptography import key_algorithm
from .asymmetric import VerifyKey


class PublicKey(VerifyKey, ABC):
    """This class is used to en/decrypt symmetric key or sign/verify signature with message data

        Asymmetric Cryptography Public Key
        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        key data format: {
            algorithm : "RSA", // "ECC", ...
            data      : "{BASE64_ENCODE}",
            ...
        }
    """

    #
    #   PublicKey factory
    #
    class Factory:

        @abstractmethod
        def parse_public_key(self, key: dict):  # -> Optional[PublicKey]:
            """
            Parse map object to key

            :param key: key info
            :return: PublicKey
            """
            raise NotImplemented

    __factories = {}

    @classmethod
    def register(cls, algorithm: str, factory: Factory):
        cls.__factories[algorithm] = factory

    @classmethod
    def factory(cls, algorithm: str) -> Optional[Factory]:
        return cls.__factories.get(algorithm)

    @classmethod
    def parse(cls, key: dict):  # -> Optional[PublicKey]:
        if key is None:
            return None
        elif isinstance(key, cls):
            return key
        elif isinstance(key, Map):
            key = key.dictionary
        algorithm = key_algorithm(key=key)
        factory = cls.factory(algorithm=algorithm)
        if factory is None:
            factory = cls.factory(algorithm='*')  # unknown
            assert factory is not None, 'cannot parse key: %s' % key
        return factory.parse_public_key(key=key)
